<!DOCTYPE html>
<html>
  {% include head.html %}
  <body class="docsNavVisible">
    {% include nav.html alwayson=true %}
    <div class="navPusher">
      <div class="docMainWrapper wrapper">
      {% include nav/collection_nav.html navdata=site.data.nav_docs type="docs" sectionpath="/docs/" sectiontitle="Docs" data_collection=site.docs %}
      {{ content }}
      </div>
      {% include footer.html %}
    </div>
  </body>
</html>
