<div class="iframeContent">
  <iframe class="pluginIframe" src="{{ include.href }}" seamless></iframe>
</div>
<div class="iframePreview">
  {% include plugins/button.html href=include.href text=include.text %}
</div>