<!DOCTYPE html>
<html>
  {% include head.html %}
  <body>
    {% include nav.html alwayson=true %}
    <div class="navPusher">
      {{ content }}
      {% include footer.html %}
    </div>
  </body>

</html>
