alter table lineitem drop column CBIGINT;
alter table lineitem drop column CDECIMAL1;
alter table lineitem drop column CDECIMAL4;
alter table lineitem drop column CDECIMAL4_2;
alter table lineitem drop column CDECIMAL5;
alter table lineitem drop column CDECIMAL9;
alter table lineitem drop column CDECIMAL9_2;
alter table lineitem drop column CDECIMAL10;
alter table lineitem drop column CDECIMAL18;
alter table lineitem drop column CDECIMAL18_2;
alter table lineitem drop column CINTEGER;
alter table lineitem drop column CSMALLINT;
alter table lineitem drop column CTINYINT;
alter table lineitem drop column CDOUBLE;
alter table lineitem drop column CFLOAT;
alter table lineitem drop column CDATE;
alter table lineitem drop column CDATETIME;
alter table lineitem drop column CTIME;
alter table lineitem drop column CCHAR1;
alter table lineitem drop column CCHAR2;
alter table lineitem drop column CCHAR3;
alter table lineitem drop column CCHAR4;
alter table lineitem drop column CCHAR5;
alter table lineitem drop column CCHAR6;
alter table lineitem drop column CCHAR7;
alter table lineitem drop column CCHAR8;
alter table lineitem drop column CCHAR9;
alter table lineitem drop column CCHAR255;
alter table lineitem drop column CVCHAR1;
alter table lineitem drop column CVCHAR2;
alter table lineitem drop column CVCHAR3;
alter table lineitem drop column CVCHAR4;
alter table lineitem drop column CVCHAR5;
alter table lineitem drop column CVCHAR6;
alter table lineitem drop column CVCHAR7;
alter table lineitem drop column CVCHAR8;
alter table lineitem drop column CVCHAR255;
alter table lineitem drop column CTEXT;
alter table lineitem drop column CTINYTEXT;
alter table lineitem drop column CMEDIUMTEXT;
alter table lineitem drop column CLONGTEXT;
alter table lineitem drop column CBLOB;
alter table lineitem drop column CTINYBLOB;
alter table lineitem drop column CMEDIUMBLOB;
alter table lineitem drop column CLONGBLOB;
