/*****************************************************************************

Copyright (c) 2000, 2009, Oracle and/or its affiliates. All Rights Reserved.

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; version 2 of the License.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1335 USA

*****************************************************************************/

/********************************************************************//**
@file include/data0types.h
Some type definitions

Created 9/21/2000 Heikki Tuuri
*************************************************************************/

#ifndef data0types_h
#define data0types_h

/* SQL data field struct */
struct dfield_t;

/* SQL data tuple struct */
struct dtuple_t;

#endif

