rpl_tokudb_delete_pk: unreliable, uses timestamp differences
rpl_tokudb_delete_pk_lookup1: unreliable, uses timestamp differences
rpl_tokudb_update_pk_uc0_lookup0: unreliable, uses timestamp differences
rpl_tokudb_update_pk_uc0_lookup1: unreliable, uses timestamp differences
rpl_tokudb_update_pk_uc1_lookup0: unreliable, uses timestamp differences
rpl_tokudb_update_pk_uc1_lookup1: unreliable, uses timestamp differences
rpl_tokudb_update_unique_uc0_lookup0: unreliable, uses timestamp differences
rpl_tokudb_update_unique_uc0_lookup1: unreliable, uses timestamp differences
rpl_tokudb_write_pk: unreliable, uses timestamp differences
rpl_tokudb_write_pk_uc1: unreliable, uses timestamp differences
rpl_tokudb_write_unique: unreliable, uses timestamp differences
rpl_tokudb_write_unique_uc1: unreliable, uses timestamp differences
rpl_tokudb_read_only_ff: unreliable, uses timestamp differences
rpl_tokudb_read_only_tf: unreliable, uses timestamp differences
rpl_tokudb_read_only_tt: unreliable, uses timestamp differences
rpl_tokudb_read_only_ft: no TOKU_INCLUDE_RFR
